import hmac
import sha
from zope.component import getUtility
from zope.interface import implements
from AccessControl import getSecurityManager
from zExceptions import Forbidden
from ZPublisher.HTTPRequest import HTTPRequest
from Products.Five import BrowserView
from plone.keyring.interfaces import IKeyManager
from plone.protect.interfaces import IAuthenticatorView
from plone.protect.utils import protect
from zope.deprecation import deprecated


def _getUserName():
    user=getSecurityManager().getUser()
    if user is None:
        return "Anonymous User"
    return user.getUserName()


def _verify(request):
    auth=request.get("_authenticator")
    if auth is None:
        return False

    manager=getUtility(IKeyManager)
    ring=manager[u"_system"]
    user=_getUserName()

    for key in ring:
        if key is None:
            continue
        correct=hmac.new(key, user, sha).hexdigest()
        if correct==auth:
            return True

    return False


class AuthenticatorView(BrowserView):
    implements(IAuthenticatorView)

    def authenticator(self):
        manager=getUtility(IKeyManager)
        secret=manager.secret()
        user=_getUserName()
        auth=hmac.new(secret, user, sha).hexdigest()
        return '<input type="hidden" name="_authenticator" value="%s"/>' % \
                auth


    def verify(self):
        return _verify(self.request)


def check(request):
    if isinstance(request, HTTPRequest):
        if not _verify(request):
            raise Forbidden('Form authenticator is invalid.')


def AuthenticateForm(callable):
    return protect(check)(callable)

deprecated("AuthenticateForm", "Please use the plone.protect.protect decorator")


__all__ = [ "AuthenticatorView", "AuthenticateForm", "check" ]




from AccessControl.requestmethod import _buildFacade
import inspect

_default = []

# This is based on AccessControl.requestmethod.postonly
# It should probably be updated to use the decorator module.
class protect(object):
    def __init__(self, *checkers):
        self.checkers = checkers

    def __call__(self, callable):
        spec = inspect.getargspec(callable)
        args, defaults = spec[0], spec[3]
        try:
            r_index = args.index("REQUEST")
        except ValueError:
            raise ValueError("No REQUEST parameter in callable signature")

        arglen = len(args)
        if defaults is not None:
            defaults = zip(args[arglen - len(defaults):], defaults)
            arglen -= len(defaults)

        def _curried(*args, **kw):
            request = None

            if len(args) > r_index:
                request = args[r_index]

            for checker in self.checkers:
                checker(request)

            # Reconstruct keyword arguments
            if defaults is not None:
                args, kwparams = args[:arglen], args[arglen:]
                for positional, (key, default) in zip(kwparams, defaults):
                    if positional is _default:
                        kw[key] = default
                    else:
                        kw[key] = positional

            return callable(*args, **kw)

        # Build a facade, with a reference to our locally-scoped _curried
        facade_globs = dict(_curried=_curried, _default=_default)
        try:
            name = callable.__name__
            exec _buildFacade(name, spec, callable.__doc__) in facade_globs
        except TypeError: # BBB: Zope 2.10
            name = '_facade'
            exec _buildFacade(spec, callable.__doc__) in facade_globs
        return facade_globs[name]




from unittest import TestCase
from unittest import TestSuite
from unittest import makeSuite
from plone.protect.utils import protect

def funcWithoutRequest():
    pass

def funcWithRequest(one, two, REQUEST=None):
    return (one, two)

class DummyChecker:
    def __call__(self, request):
        self.request=request

class DecoratorTests(TestCase):
    def testFunctionMustHaveRequest(self):
        protector=protect()
        self.assertRaises(ValueError, protector, funcWithoutRequest)

    def testArgumentsPassed(self):
        wrapped=protect()(funcWithRequest)
        self.assertEqual(wrapped("one", "two"), ("one", "two"))

    def testKeywordArguments(self):
        wrapped=protect()(funcWithRequest)
        self.assertEqual(wrapped(one="one", two="two"), ("one", "two"))

    def testMixedArguments(self):
        wrapped=protect()(funcWithRequest)
        self.assertEqual(wrapped("one", two="two"), ("one", "two"))

    def testRequestPassedToChecker(self):
        checker=DummyChecker()
        wrapped=protect(checker)(funcWithRequest)
        request=[]
        wrapped("one", "two", request)
        self.failUnless(checker.request is request)



def test_suite():
    suite=TestSuite()
    suite.addTest(makeSuite(DecoratorTests))
    return suite



from unittest import TestCase
from zope.component import getGlobalSiteManager
from zope.interface import implements
from plone.keyring.interfaces import IKeyManager
from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManagement import noSecurityManager
from AccessControl.User import User


class MockKeyManager:
    implements(IKeyManager)

    keys = [ "one", "two", "three", "four", "five" ]

    def secret(self):
        return self.keys[0]

    def __getitem__(self, key):
        return self.keys


class KeyringTestCase(TestCase):
    def setUp(self):
        self.sm=getGlobalSiteManager()
        self.manager=MockKeyManager()
        self.sm.registerUtility(self.manager, provided=IKeyManager, event=False)
        # Tests modify the user object so we better make sure it is *our*
        # user object and not the built-in Anonymous User.
        newSecurityManager(None, User('dummy', 'secret', (), ()))

    def tearDown(self):
        self.sm.unregisterUtility(self.manager, provided=IKeyManager)
        noSecurityManager()




from zope.interface import Interface

class IAuthenticatorView(Interface):

    def authenticator():
        """Return an xhtml snippet which sets an authenticator.
        
        This must be included inside a <form> element.
        """

    def verify():
        """
        Verify if the request contains a valid authenticator.
        """



import hmac
import sha
import sys
from unittest import TestSuite
from unittest import makeSuite
from AccessControl import getSecurityManager
from zExceptions import Forbidden
from ZPublisher.HTTPRequest import HTTPRequest
from plone.protect.tests.case import KeyringTestCase
from plone.protect.authenticator import AuthenticatorView
from plone.protect.authenticator import AuthenticateForm



class AuthenticatorTests(KeyringTestCase):
    def setUp(self):
        KeyringTestCase.setUp(self)
        self.view=AuthenticatorView(None, None)


    def setUsername(self, name):
        user=getSecurityManager().getUser()
        user.name=name


    def setSecret(self, secret):
        self.manager.keys[0]=secret


    def testIsHtmlInput(self):
        auth=self.view.authenticator()
        self.failUnless(auth.startswith("<input"))
        self.failUnless(auth.endswith("/>"))


    def testConsistent(self):
        one=self.view.authenticator()
        two=self.view.authenticator()
        self.assertEqual(one, two)


    def testDiffersPerUser(self):
        one=self.view.authenticator()
        self.setUsername("other")
        two=self.view.authenticator()
        self.assertNotEqual(one, two)


    def testDiffersPerSecret(self):
        one=self.view.authenticator()
        self.setSecret("other")
        two=self.view.authenticator()
        self.assertNotEqual(one, two)



class VerifyTests(KeyringTestCase):
    def setUp(self):
        self.request={}
        KeyringTestCase.setUp(self)
        self.view=AuthenticatorView(None, self.request)


    def setAuthenticator(self, key):
        user=getSecurityManager().getUser().getUserName()
        auth=hmac.new(key, user, sha).hexdigest()
        self.request["_authenticator"]=auth


    def testCorrectAuthenticator(self):
        self.manager.keys[0]=("secret")
        self.setAuthenticator("secret")
        self.assertEqual(self.view.verify(), True)


    def testOlderSecretVerifies(self):
        self.manager.keys[3]="backup"
        self.setAuthenticator("backup")
        self.assertEqual(self.view.verify(), True)


    def testMissingAuthenticator(self):
        self.assertEqual(self.view.verify(), False)


    def testIncorrectAuthenticator(self):
        self.request["_authenticator"]="incorrect"
        self.assertEqual(self.view.verify(), False)


    def testAuthenticatorWrongType(self):
        self.request["_authenticator"]=123
        self.assertEqual(self.view.verify(), False)



class DecoratorTests(KeyringTestCase):
    def setUp(self):
        self.request=HTTPRequest(sys.stdin, dict(SERVER_URL="dummy"), None)
        KeyringTestCase.setUp(self)
        def func(REQUEST=None):
            return 1
        self.func=AuthenticateForm(func)


    def testNoRequestParameter(self):
        def func():
            pass
        self.assertRaises(ValueError, AuthenticateForm, func)


    def testIgnoreBadRequestType(self):
        self.assertEqual(self.func(), 1)


    def testBadAuthenticator(self):
        self.request["_authenticator"]="incorrect"
        self.assertRaises(Forbidden, self.func, self.request)


def test_suite():
    suite=TestSuite()
    suite.addTest(makeSuite(AuthenticatorTests))
    suite.addTest(makeSuite(VerifyTests))
    suite.addTest(makeSuite(DecoratorTests))
    return suite



import os
from setuptools import setup, find_packages

version = '1.0'

setup(name='plone.protect',
      version=version,
      description="Security for browser forms",
      long_description=open("README.txt").read() + "\n" + \
                       open(os.path.join("docs", "HISTORY.txt")).read(),
      classifiers=[
        "Framework :: Plone",
        "Framework :: Zope3",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='zope security CSFS',
      author='Wichert Akkerman',
      author_email='plone-developers@lists.sourceforge.net',
      url='http://svn.plone.org/svn/plone/plone.protect',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['plone', ],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          "plone.keyring",
      ],
      )
